/*******************************************************************/
/*                                                                 */
/*   File name :  libvdif.h                                        */
/*                                                                 */
/*      Version:   2014-04-23                                      */
/*                                                                 */
/*   Copyright (c) 2014-     T.Kondo/NICT All Right Reserved       */
/*                                                                 */
/*******************************************************************/
#ifndef _LIBVDIF_
#define _LIBVDIF_

#include <stdio.h>
#include <signal.h> 
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "lib_from_K5.h"

#ifdef __CYGWIN__
#ifndef linux
#define linux
#endif
#endif

#ifdef linux
#include <sys/time.h>
#endif

#ifndef WIN32
#include <unistd.h>
#endif



#ifdef linux
 #ifdef _K5_USE_FOPEN64_
  #ifndef _FOPEN_DEFINED_
   #define _FOPEN_DEFINED_
   #define fopen(_Z,_X) fopen64(_Z,_X)
  #endif
 #endif
 #ifdef _LARGEFILE_SOURCE
  #ifndef _FSEEK_DEFINED_
   #define _FSEEK_DEFINED_
   #define fseek(_Z,_X,_W) fseeko(_Z,_X,_W)
  #endif
  #ifndef _FTELL_DEFINED_
   #define _FTELL_DEFINED_
   #define ftell(_Z) ftello(_Z)
  #endif
 #endif
#endif


/* for WIN32 */

#ifdef WIN32
#pragma warning (disable:4996)
#ifndef _FORWIN32SET_
 #define _FORWIN32SET_
 typedef unsigned int    u_int32_t;
 typedef unsigned int    uint32_t;
 typedef unsigned short int    u_int16_t;
 typedef unsigned short int    uint16_t;
 typedef unsigned char   u_int8_t;
 typedef _int64 int64_t;
 #define popen(_Z,_X) _popen(_Z,_X)
 #define pclose(_Z) _pclose(_Z)
 #define fseek(_Z,_X,_W) _fseeki64(_Z,_X,_W)
 #define ftell(_Z) _ftelli64(_Z)
 #endif
#endif


#ifndef USE_ASM
#define USE_ASM
#endif


/* constant set */
#ifndef _SET_K5_CONSTANTS_
#define _SET_K5_CONSTANTS_

#define HEAD_SEARCH_RANGE 600  // header serach range (in 32bit word) -0.5*HEAD_SEARCH_RANGE to +0.5*HEAD_SEARCH_RANGE
#define VDIF_KASHIMA_SYNC 0xACABFEEDUL    // sync pattern for VDIF Kashima format

#ifndef FILEMAXSIZE
#define FILEMAXSIZE  2000000000   //  File maximum limit size in bytes when limiting
#endif

#ifndef PIRA
#define PIRA   0.017453292519943295769236907684886    /* PI/180  degree to radian */
#endif
#ifndef PIDE
#define PIDE   57.295779513082320876798154814105      /* 180/PI radian to degree  */
#endif
#ifndef TWOPI
#define TWOPI  6.283185307179586476925286766559       /* 2PI                      */
#endif
#ifndef PI
#define PI     3.1415926535897932384626433832795      /* PI                       */
#endif
#ifndef HALFPI
#define HALFPI 1.5707963267948966192313216916398      /*  PI/2                    */
#endif
#ifndef VLIGHT
#define VLIGHT 2.99792458E+8    /* speed of light (m/sec)  */
#endif

/* mainly used in libapri etc */
#ifndef CONVA
#define CONVA 4.8481368110953599358991410235795e-6    /* PI/(3600*180)  arcsec to radian */
#endif
#ifndef CONVS
#define CONVS 7.2722052166430399038487115353692e-5    /* PI/(3600*12) time sec to radian */
#endif
#ifndef CONVD
#define CONVD 0.017453292519943295769236907684886    //*PI/180  degree to radian  */
#endif
#ifndef CONVH
#define CONVH 0.26179938779914943653855361527329    /* PI/12 hour to radian  */
#endif


#define REWIND_BYTE 200       // for K5/VSSP access

#endif   // *** END of _SET_K5_CONSTANTS_

typedef struct timeval_vdif    // time related variables used in vdif utilities
{
	int sec_fm_ref;   // Seconds from reference epoch
	int ref_epo;      // Reference epoch for second count.  

    int year;      // 4 digit year
    int ddd;       // day of year
    int hh;        // hour
    int mm;        // minute
    int ss;        // sec in minute
    int second;    // sec in day
    double dtime;  // time in day (ddd+second/86400.0)
    int month;     // month
    int mday;      // day of month
	int mjd;       // MJD
	double fday;   // fractional part of day
} timeval_vdif_t;



typedef struct vdif_header     // vdif file header structure
{
	int sec_fm_ref;   // Seconds from reference epoch
	int legacy;       // Legacy mode  0: standard 32 byte VDIF data frame header
	                  //              1: legacy header-length' mode ; words 4-7 omitted from header
	int invalid;      // Invalid data   0: valid  1: invalid
	int frame;        // Data frame # within second, starting at zero; must be integral number od data drames per second
	int ref_epo;      // Reference epoch for second count.  This contains the 6-month period in which the VDIF
	                  // clock was set, with an origin of at 00UTC 1 Jan 2000, such that '0' corresponds to the
	                  // first 6 months of 2000, '1' corresponds to the 6 months starting 00h 1 July 2000, etc. 
	                  // This field rolls over to 0 again at the 6-month period starting 00h 1 Jan 2032.
	int frame_len64;    // Data fram length (including header) in units of 8 bytes (must be multiple of 8 bytes, 
	                  // max is 2^27 bytes)
	int log2_chs;     // log_2(#channels in data array). #chans must be power of 2
	int ver;           // VDIF version number
	unsigned short station_id;   //  station ID (either A2 of unsigned 16-bit integer)
	char sid[3];      // station ID (A2)
	int thread_id;    // Thread ID (0-1023)
	int bpsampl;      // bits per sample (max 32). If complex type, this is set according to the #bits in each complex-
	                  // component
	int d_type;       // data type  0: real data   1: complex data
	int edv;          // Extended data version
	int w1;           // extented data word#1 (3 bytes are valid)
	int w2;           // extended data word#2
	int w3;           // extended data word#3
	int w4;           // extended data word#4
	// below is extended format of NICT's
	int uflag;        // unit of sampling rate 0: kHz,  1: MHz
	int srate;        // sampling rate
	int sync_block;   // sync block  = w2
	char das[9];      // DAS/station name1
    timeval_vdif_t t; // time information converted to YYYY Month Day hour minute second

} vdif_header_t;

typedef struct ch_data     // ch_data variable definition
{
    int *idata;
    int idata_over[32];
    int usampl;
} ch_data_t;

typedef struct vdif_file     // vdif_file type variable definition
{
    /*
        Ver. 2014-04-25
         and previous one.
    */
	char fname[128]; //data file name opened or to be opened  **K5**
	FILE *fin;
	int64_t skip_bytes;   // skip bytes (1st skip over # of bytes)
	int64_t pos;   // current file pointer
	int64_t bytes1sec;  // # of bytes in 1 second including headder portion
	                    // (usually used for 1 st forward)
	int64_t f_size;  // opened file size in bytes
	int64_t pos_1st_header;  // position of 1 st correct header (valid when flag_vdif_anal=1)
	int64_t tot_frame; // total frame counter
	//int64_t tbytesread;   // total # of bytes read
    vdif_header_t vhead;
	int flag_vdif_anal;  //  flag to show vdif file format is analyzed
	double tnframe;   // total number of frames (double) (computed from file size)
    double fbps;      // supposed data rate (bps) per thread
	double fsample;   // sampling frequency (Hz)
	double tsample;   // sampling period (sec)
	double dlength;   // data length in seconds
	int headlen;      // header length in 32bit word
	int nframsec;     // frames/sec per thread
	int adbit;      // AD resolution
	int numch;  // # of channels
	int nthread; // # of threads
	int frame_bytes; // # of bytes in a frame (excepting header)
	double unitp;   // usampl*tsampl unit length in sec
    u_int32_t head[8];      // for reading header portion
    u_int32_t *ubuf;        // for data field reading
    int nword;       // ubuf size (in 32bit word) allocated or to be allocated
	int nsample;     // # of samples in ubuf
	int imax;        // # of ubuf read for 1 frame
	int imax_c;      // counter to compare with imax (already read number)
	int upos;  // next read (decode) position in ubuf array
	int wordread;     // actually read words (ubuf)
    //int *idata;        // pointer to decoded data (integer) container by channel
    //double *ddata;      // pointer to decoded data (R*8) container
	//int idata_over[32]; // to contain overflowed decode data
    ch_data_t *ch;   // decode data by channel (max numch)
	int smode;     // search mode
	int usampl;      // usampl*numch is the size of idata and data
	int usampl_c;    // counter used for actual number of samples decoded data 
	int n_usampl;    // # of unit samples (starting from 1) 
	double dtime;  // time at begin of unit
	double dtimest;  // time at begin of file data  (in day)
	int npos;    //  valid number in idata or data
	int flag_idata_over;  // flag to show how many samples are over from idata  0: ok  +ve: overed # 
	int flag_read_data;      // data read flag
	int flag_read_header;   // header read flag
	int flag_unit_read_ok;  // 1 unit data read OK flag
	int flag_compulsory;  // 1: set header format to Kashima compulsory (debug purpose)
	                      // +VE (>1) : set bytes/frame compulsory with Kashima Format 
	int flag_open;        // file open flag  0: closed  1:opened
	//int flag_elec;       // set ELEC VDIF format compulsory (debug purpose)

/*
	int icnt;   //  current unit number in a second data 
	int numb;   //  # of bytes in unit # of samples
	int usampl;  // unit # of samples
	int kpos;  // should be obsoleted ?
	int imax;    //  # of units in a second data
	int bytesec;    //  # of bytes in a second data  **K5**
	//char* fname; //pointer data file name obsoleted
	char fname[128]; //data file name opened or to be opened  **K5**
	double fsample;  // sampling frequency (Hz)
	double tsample; // sampling period (sec)
	double unitp;   // usampl*tsampl unit length 
    //long int seconds; // time (sec) from mid night (00h00m00s)
    int seconds; // time (sec) from mid night (00h00m00s)  2011-02-07
	int hh, mm, ss;  // time in hour, minutes, seconds 
    int eflg; // error flag (K5/VSSP32)  added on 2007-02-07
    int year; // 4 digits year (K5/VSSP32)
    int totalday; // (K5/VSSP32)
    int version1; // Major version (K5/VSSP32)
    int version2; // Minor version (K5/VSSP32)
	int spatn; // detected 2nd sync pattern detected 
	int syncok; // =1 for 32bit all 1 detection // not used
	int fh;
	int ch;    
	int ierr;   // error code (general purpose)
	double dtime;  // time at begin of unit
	double dtimest;  // time at begin of file data
	unsigned int bytesread;  // number of bytes read
	int64_t tbytesread;   // total # of bytes read **K5**
	int wordread;   // # of 32bit-words read counter (general) **K5**
	int64_t twordread;   // total # of 32bit-word read **K5**
    int vflag, vflag_old;  // byte slip flag  added on 2003-02-14
                           //  if header sync detection is suceed
                           //     set to 1.  else 0
                           //   1: sync detection success 
                           //   0: sync detection error (may be byte slip)
                           //   vflag_old is the previous status
    //long int dtime_vflag, dtime_vflag_old; // time at vflag decision made
    int dtime_vflag, dtime_vflag_old; // time at vflag decision made  2011-02-07
	// add 2007-03-02 for robust sync pattern dfinding
	//      sampling parameters and time at 1 st frame
	int sfreq1st;  //  sampling frequency in kHz
	int adbit1st;  // AD bits (1,2,4,8)
	int numch1st;  // # of channels
    //long int seconds1st; // time (sec) from mid night (00h00m00s)
    //long int secold; // time (sec) from mid night (00h00m00s)
    long int seconds1st; // time (sec) from mid night (00h00m00s)  2011-02-07
    long int secold; // time (sec) from mid night (00h00m00s)      2011-02-07
	int framenum;  // frame number starting from 0
*/
} vdif_file_t;


#ifdef __cplusplus
extern "C"{
#endif 

void parse_vdif_header(int *head8, vdif_header_t *vhead);
void ut2mjd_v(timeval_vdif_t *t);
void mjd2ut_v(timeval_vdif_t *t);
void vdiftime2ymdhms(timeval_vdif_t *t);
int check_vdif_header(FILE *fin);
int find_Kas_sync(FILE *fin, int64_t *pos, int mode);
int get_real_vdif_info(vdif_file_t *vd, int mode);

int vdif_t_ini(vdif_file_t *vd, int mode);
int fwdNsecVDIF(int nsec, vdif_file_t *vd);
int vdif_header_read_mode1(vdif_file_t *vd);
int vdif_header_read_mode2(vdif_file_t *vd);
int vdif_header_read_check(vdif_file_t *vd, int mode, int64_t *sa_pos);
int vdif_read_unit(vdif_file_t *vd, int r_thread_id);

#ifdef __cplusplus
}
#endif

#endif
